from bpc8583.tools import get_random_hex

class Card:
	def __init__(self, pan=None, expiry_date=None, service_code=None, pvvki=None, PVV=None, CVV=None, discretionary_data=None, description=None):
		"""
		"""
		self.trxn_counter = 0
		self.set_description(description)

		if pan:
			self.pan = str(pan)
		else:
			self.pan = '8990011234567890'

		if expiry_date:
			self.expiry_date = str(expiry_date)
		else:
			self.expiry_date = '1809'

		if service_code:
			self.service_code = str(service_code)
		else:
			self.service_code = '201'

		if pvvki:
			self.PVV_key_index = str(pvvki)
		else:
			self.PVV_key_index = '1'

		if PVV:
			self.PVV = str(PVV)
		else:
			self.PVV = '8723'

		if CVV:
			self.CVV = str(CVV)
		else:
			self.CVV = '000'

		if discretionary_data:
			self.discretionary_data = str(discretionary_data)
		else:
			self.discretionary_data = ''

	def get_iss_application_data(self):
		"""
		Get ICC Issuer Application data (Contains proprietary application data for transmission to the issuer in an online transaction)
		"""
		return '0000'


	def get_application_cryptogram(self):
		"""
		Return ICC application crytpogram (returned by the ICC in response of the GENERATE AC command)
		"""
		self.trxn_counter += 1
		return get_random_hex(16)


	def _set_transaction_counter(self, counter):
		"""
		"""
		self.trxn_counter = counter


	def get_transaction_counter(self):
		"""
		Get ICC application transaction counter
		"""
		return str(self.trxn_counter).rjust(4, '0')


	def get_card_number(self):
		"""
		"""
		return self.get_int_card_number()


	def get_service_code(self):
		"""
		"""
		return self.service_code


	def get_int_card_number(self):
		"""
		"""
		return int(self.pan)


	def get_str_card_number(self):
		"""
		"""
		return self.pan


	def get_expiry_date(self):
		"""
		"""
		return int(self.expiry_date)


	def get_track2(self):
		"""
		"""
		return self.pan + '=' + self.expiry_date + self.service_code + self.PVV_key_index + self.PVV + self.CVV + self.discretionary_data


	def get_sequence_number(self):
		"""
		"""
		return 1

	def set_description(self, description):
		"""
		"""
		if description: 
			self.description = description
		else:
			self.description = ''


	def get_description(self):
		"""
		"""
		return self.description




import struct
from datetime import datetime

from bpc8583.ISO8583 import ISO8583, currency_codes
from bpc8583.tools import get_date, get_datetime_with_year, get_datetime, get_stan, get_seconds_since_epoch, get_MMDD
from tracetools.tracetools import trace
from bpc8583.spec import IsoSpec, IsoSpec1987BPC
from pytlv.TLV import TLV

class Transaction():
    def __init__(self, type, card, term, icc_trxn=None):
        """
        """
        self.TLV = TLV()

        self.IsoMessage = ISO8583(IsoSpec=IsoSpec1987BPC())
        self.card = card
        self.term = term
        self.type = type.lower()
        self.description = ''
        self.expected_response_code = '000'
        self.expected_response_action = None
        self.currency = None
        self.PIN = None
        self._set_icc_trxn(icc_trxn)

        if self.type in ['logon', 'echo']:
            """
            """
            self.IsoMessage.MTI("0800")

            self.IsoMessage.FieldData(3, 990000)
            self.IsoMessage.FieldData(12, get_datetime_with_year())
            self.IsoMessage.FieldData(24, 801)

        elif self.type == 'key change':
            """
            """
            self.IsoMessage.MTI("0800")

            self.IsoMessage.FieldData(3, 990000)
            self.IsoMessage.FieldData(12, get_datetime_with_year())
            self.IsoMessage.FieldData(24, 811)

        elif self.type == 'balance':
            """
            """
            self.IsoMessage.MTI("0100")
            
            self.IsoMessage.FieldData(2, self.card.get_card_number())
            self.IsoMessage.FieldData(3, 310000)
            #self.IsoMessage.FieldData(4, 0)
            self.IsoMessage.FieldData(12, get_datetime_with_year())
            self.IsoMessage.FieldData(13, get_MMDD())
            self.IsoMessage.FieldData(22, self.term.get_pos_entry_mode())
            self.IsoMessage.FieldData(24, 100)
            self.IsoMessage.FieldData(25, 0)
            self.IsoMessage.FieldData(35, self.card.get_track2())

        elif self.type == 'purchase':
            """
            """
            self.IsoMessage = ISO8583(IsoSpec=IsoSpec1987BPC())            
            self.IsoMessage.MTI("0100")
        
            self.IsoMessage.FieldData(2, self.card.get_card_number())
            self.IsoMessage.FieldData(3, 000000)
            self.IsoMessage.FieldData(12, get_datetime_with_year())
            self.IsoMessage.FieldData(22, self.term.get_pos_entry_mode())
            self.IsoMessage.FieldData(24, 100)
            self.IsoMessage.FieldData(25, 0)
            self.IsoMessage.FieldData(35, self.card.get_track2())

        elif self.type == 'refund':
            """
            Refund
            """
            self.IsoMessage = ISO8583(IsoSpec=IsoSpec1987BPC())
            self.IsoMessage.MTI("0100")
        
            self.IsoMessage.FieldData(2, self.card.get_card_number())
            self.IsoMessage.FieldData(3, 200000)
            self.IsoMessage.FieldData(12, get_datetime_with_year())
            self.IsoMessage.FieldData(22, self.term.get_pos_entry_mode())
            self.IsoMessage.FieldData(24, 100)
            self.IsoMessage.FieldData(25, 0)
            self.IsoMessage.FieldData(35, self.card.get_track2())

        elif self.type == 'cash':
            """
            Refund
            """
            self.IsoMessage = ISO8583(IsoSpec=IsoSpec1987BPC())
            self.IsoMessage.MTI("0100")
        
            self.IsoMessage.FieldData(2, self.card.get_card_number())
            self.IsoMessage.FieldData(3, 120000)
            self.IsoMessage.FieldData(12, get_datetime_with_year())
            self.IsoMessage.FieldData(22, self.term.get_pos_entry_mode())
            self.IsoMessage.FieldData(24, 100)
            self.IsoMessage.FieldData(25, 0)
            self.IsoMessage.FieldData(35, self.card.get_track2())

        else:
            print('Unknown transaction type: {}'.format(type))
            return None
    
        # Common message fields:
        self.IsoMessage.FieldData(7, get_seconds_since_epoch())
        self.IsoMessage.FieldData(11, get_stan())
        self.IsoMessage.FieldData(41, self.term.get_terminal_id())
        self.IsoMessage.FieldData(42, self.term.get_merchant_id())
        self.IsoMessage.FieldData(49, self.term.get_currency_code())

        if self.type in ['purchase', 'balance', 'cash'] and self.icc_trxn:
            self.IsoMessage.FieldData(55, self.build_emv_data())

        self.rebuild()


    def get_data(self):
        """
        """
        return struct.pack("!H", len(self.data)) + self.data


    def rebuild(self):
        """
        Rebuild IsoMessage (e.g. after field data change)
        """
        self.data = self.IsoMessage.BuildIso()


    def trace(self, header=None):
        """
        """
        self.IsoMessage.Print(header=header)


    def set_description(self, description):
        """
        """
        self.description = description


    def get_description(self):
        """
        Get transaction description (for logging purposes)
        """
        card_description = self.card.get_description()
        if card_description:
            card_description += ' | '

        if self.description:
            return card_description + self.description
        else:
            return card_description + self.type + ' ' + str(self.IsoMessage.FieldData(11))


    def set_PIN(self, PIN):
        """
        """
        if PIN:
            self.PIN = PIN
            encrypted_pinblock = self.term.get_encrypted_pin(PIN, self.card.get_card_number())
            if encrypted_pinblock:
                self.IsoMessage.FieldData(52, encrypted_pinblock)
                self.rebuild()

    def get_PIN(self):
        """
        """
        return self.PIN


    def set_amount(self, amount):
        """
        Set transaction amount
        """
        if amount:
            try:
                self.IsoMessage.FieldData(4, int(amount))
            except ValueError:
                self.IsoMessage.FieldData(4, 0)
            self.rebuild()


    def set_expected_code(self, expected_response_code):
        """
        Expected response code of the transaction
        """
        self.expected_response_code = expected_response_code


    def set_expected_action(self, expected_response_action):
        """
        Expected outcome of the transaction ('APPROVED' or 'DECLINED')
        """
        if expected_response_action.upper() not in ['APPROVED', 'APPROVE', 'DECLINED', 'DECLINE']:
            return False

        self.expected_response_action = expected_response_action.upper()
        return True


    def is_response_expected(self, actual_response_code):
        """
        """
        if self.expected_response_action in ['APPROVED', 'APPROVE']:
            if int(actual_response_code) == 0:
                return True
            else:
                return False

        elif self.expected_response_action in ['DECLINED', 'DECLINE']:
            if int(actual_response_code) != 0:
                return True
            else:
                return False
        else:
            # no response action available, compare the response codes:
            if self.expected_response_code:
                if self.expected_response_code == actual_response_code:
                    return True
                else:
                    return False
            else:
                # neither response action, nor response code are set
                return None

        return True


    def _get_app_interchange_profile(self):
        """
        Contains the data objects (with tags and lengths) returned by the ICC in response to a command
        """
        #return '9A039505'
        return '0000'


    def build_emv_data(self):
        """
        TODO:
        95    TVR
        82    app_int_prof 
        """
        emv_data = ''
        emv_data += self.TLV.build({'82': self._get_app_interchange_profile()})
        emv_data += self.TLV.build({'9A': get_date()})
        emv_data += self.TLV.build({'95': self.term.get_tvr()})
        emv_data += self.TLV.build({'9F10': self.card.get_iss_application_data()})
        emv_data += self.TLV.build({'9F26': self.card.get_application_cryptogram()})
        emv_data += self.TLV.build({'9F36': self.card.get_transaction_counter()})
        emv_data += self.TLV.build({'9F37': self.term.get_unpredno()})
        emv_data += self.TLV.build({'9F1A': self.term.get_country_code()})
        return emv_data


    def _set_icc_trxn(self, icc_trxn):
        """
        """
        if icc_trxn == None and self.card.get_service_code()[0] in ['2', '6']:
            self.icc_trxn = True
        elif icc_trxn == None and self.card.get_service_code()[0] not in ['2', '6']:
            self.icc_trxn = False
        elif icc_trxn.lower() == 'true':
            self.icc_trxn = True
        else:
            self.icc_trxn = False


    def set_currency(self, currency_id):
        """
        Set transaction currency code from given currency id, e.g. set 840 from 'USD'
        """
        try:
            self.currency = currency_codes[currency_id]
            self.IsoMessage.FieldData(49, self.currency)
            self.rebuild()
        except KeyError:
            self.currency = None


    def get_currency(self):
        """
        Currency getter
        """
        return self.currency



from .ISO8583 import ISO8583, DT, LT, SpecError
     
class IsoSpec(object):
    __ValidContentTypes = ('a', 'n', 's', 'an', 'as', 'ns', 'ans', 'b', 'z')
    
    Descriptions = {}
    ContentTypes = {}
    DataTypes = {}
    
    def __init__(self):
        self.SetDescriptions()
        self.SetContentTypes()
        self.SetDataTypes()
    
    def SetDescriptions(self):
        pass
    def SetContentTypes(self):
        pass
    def SetDataTypes(self):
        pass
         
    def Description(self, field, Description = None):
        if(Description == None):
            return self.Descriptions[field]
        else:
            self.Descriptions[field] = Description

    def DataType(self, field, DataType = None):
        if(DataType == None):
            return self.DataTypes[field]['Data']
        else:
            if(DataType not in DT):
                raise SpecError("Cannot set data type '{0}' for F{1}: Invalid data type".format(DataType, field))
            if(field not in self.DataTypes.keys()):
                self.DataTypes[field] = {}
            self.DataTypes[field]['Data'] = DataType
    
    def ContentType(self, field, ContentType = None):
        if(ContentType == None):
            return self.ContentTypes[field]['ContentType']
        else:
            if(ContentType not in self.__ValidContentTypes):
                raise SpecError("Cannot set Content type '{0}' for F{1}: Invalid content type".format(ContentType, field))
            self.ContentTypes[field]['ContentType'] = ContentType
            
    def MaxLength(self, field, MaxLength = None):
        if(MaxLength == None):
            return self.ContentTypes[field]['MaxLen']
        else:
            self.ContentTypes[field]['MaxLen'] = MaxLength
    
    def LengthType(self, field, LengthType = None):
        if(LengthType == None):
            return self.ContentTypes[field]['LenType']
        else:
            if(LengthType not in self.__ValidContentTypes):
                raise SpecError("Cannot set Length type '{0}' for F{1}: Invalid length type".format(LengthType, field))
            self.ContentTypes[field]['LenType'] = LengthType
    
    def LengthDataType(self, field, LengthDataType = None):
        if(LengthDataType == None):
            return self.DataTypes[field]['Length']
        else:
            if(LengthDataType not in DT):
                raise SpecError("Cannot set data type '{0}' for F{1}: Invalid data type".format(LengthDataType, field))
            if(field not in self.DataTypes.keys()):
                self.DataTypes[field] = {}
            self.DataTypes[field]['Length'] = LengthDataType
    
    
    
class IsoSpec1987(IsoSpec):
    def SetDescriptions(self):
        self.Descriptions = Descriptions['1987']
    def SetContentTypes(self):
        self.ContentTypes = ContentTypes['1987']

        
class IsoSpec1987ASCII(IsoSpec1987):
    def SetDataTypes(self):
        self.DataType('MTI', DT.ASCII)
        self.DataType(1, DT.ASCII) # bitmap
        
        for field in self.ContentTypes.keys():
            self.DataType(field, DT.ASCII)
            if(self.LengthType(field) != LT.FIXED):
                self.LengthDataType(field, DT.ASCII)


class IsoSpec1987BPC(IsoSpec1987):
    """
    A BPC's flavour of ISO8583 

    MTI is ASCII
    Bitmap is BCD
    DataFields are also ASCII
    """
    RespCodeDescriptions = {
        '00': 'APPROVED',
        '000': 'APPROVED',
        '001': 'HONR W/ID:',
        '005': 'UNABLE TO PROCESS',
        '020': 'BALANCE IS NEGATIVE',
        '100': 'CARD DECLINED',
        '101': 'EXPIRED CARD',
        '103': 'CALL ISSUER',
        '104': 'CAPTURE CARD SPEC',
        '106': 'PIN-TRIES EXCEED',
        '107': 'CALL ISSUER',
        '109': 'INVALID TERMINAL OR MERCHANT',
        '110': 'INVALID AMOUNT',
        '111': 'INVALID CARD',
        '116': 'INSUFFICIENT FUNDS',
        '117': 'INCORRECT PIN',
        '119': 'SECURITY VIOLATION',
        '120': 'SECURITY VIOLATION',
        '121': 'EXCDS WDRWL LIMT',
        '123': 'EXCDS WDRWL LIMT',
        '125': 'INVALID CARD',
        '202': 'INVALID CARD',
        '203': 'PICK-UP CARD',
        '204': 'PICK-UP CARD',
        '208': 'PICK-UP CARD',
        '209': 'PICK-UP CARD',
        '902': 'INVALID TRANS',
        '903': 'RE-ENTER TRANS',
        '904': 'FORMAT ERROR',
        '907': 'HOST NOT AVAIL.',
        '909': 'INVALID TRANS',
        '910': 'HOST NOT AVAIL.',
        '913': 'INVALID TRANS',
        '914': 'ORIG TRANS NOT FOUND',
        '920': 'PIN ERROR',
        '940': 'CAPTURE CARD SPEC',
    }

    def SetDataTypes(self):
        """
        """
        self.DataType('MTI', DT.ASCII)
        for field in self.ContentTypes.keys():
            self.DataType(field, DT.ASCII)
            if(self.LengthType(field) != LT.FIXED):
                self.LengthDataType(field, DT.ASCII)

        self.DataType(1, DT.BIN) # bitmap
        self.DataType(52, DT.BIN) # pin block
        self.DataType(53, DT.BIN) # security related control information
        self.DataType(55, DT.BIN) # ICC data


    def RespCodeDescription(self, response_code):
        """
        """
        try:
            return self.RespCodeDescriptions[response_code ]
        except KeyError:
            return ''


class IsoSpec1987BCD(IsoSpec1987):
    def SetDataTypes(self):
        self.DataType('MTI', DT.BCD)
        self.DataType(1, DT.BIN) # bitmap
        
        # Most popular BCD implementations use the reserved/private fields
        # as binary, so we have to set them as such in contrast to the ISO spec
        for field in self.ContentTypes.keys():
            if(self.MaxLength(field) == 999):
                self.ContentType(field, 'b')
        
        
        for field in self.ContentTypes.keys():
            
            ContentType = self.ContentType(field)
            
            if('a' in ContentType or 's' in ContentType):
                self.DataType(field, DT.ASCII)
            elif(ContentType == 'b'):
                self.DataType(field, DT.BIN)
            else:
                self.DataType(field, DT.BCD)

            if(self.LengthType(field) != LT.FIXED):
                self.LengthDataType(field, DT.BCD)

Descriptions = {}

Descriptions['1987'] = {
    1 : 'Bitmap' ,
    2 : 'Primary account number (PAN)' ,
    3 : 'Processing code' ,
    4 : 'Amount, transaction' ,
    5 : 'Amount, settlement' ,
    6 : 'Amount, cardholder billing' ,
    7 : 'Transmission date & time (MMDDhhmmss)' ,
    8 : 'Amount, cardholder billing fee' ,
    9 : 'Conversion rate, settlement' ,
    10 : 'Conversion rate, cardholder billing' ,
    11 : 'System trace audit number' ,
    12 : 'Time, local transaction (YYMMDDhhmmss)' ,
    13 : 'Date, local transaction (MMDD)' ,
    14 : 'Date, expiration' ,
    15 : 'Date, settlement' ,
    16 : 'Date, conversion' ,
    17 : 'Date, capture' ,
    18 : 'Merchant type' ,
    19 : 'Acquiring institution country code' ,
    20 : 'PAN extended, country code' ,
    21 : 'Forwarding institution. country code' ,
    22 : 'Point of service entry mode' ,
    23 : 'Application PAN sequence number' ,
    24 : 'Network International identifier (NII)' ,
    25 : 'Point of service condition code' ,
    26 : 'Point of service capture code' ,
    27 : 'Authorizing identification response length' ,
    28 : 'Amount, transaction fee' ,
    29 : 'Amount, settlement fee' ,
    30 : 'Amount, transaction processing fee' ,
    31 : 'Amount, settlement processing fee' ,
    32 : 'Acquiring institution identification code' ,
    33 : 'Forwarding institution identification code' ,
    34 : 'Primary account number, extended' ,
    35 : 'Track 2 data' ,
    36 : 'Track 3 data' ,
    37 : 'Retrieval reference number' ,
    38 : 'Authorization identification response' ,
    39 : 'Response code' ,
    40 : 'Service restriction code' ,
    41 : 'Terminal ID' ,
    42 : 'Merchant number' ,
    43 : 'Card acceptor name/location' ,
    44 : 'Additional response data' ,
    45 : 'Track 1 data' ,
    46 : 'Amount, fees' ,
    47 : 'Additional data' ,
    48 : 'Additional data - private' ,
    49 : 'Currency code, transaction' ,
    50 : 'Currency code, settlement' ,
    51 : 'Currency code, cardholder billing' ,
    52 : 'PIN data' ,
    53 : 'Security related control information' ,
    54 : 'Additional amounts' ,
    55 : 'ICC Sys Related Data' ,
    56 : 'Reserved ISO' ,
    57 : 'Reserved national' ,
    58 : 'Reserved national' ,
    59 : 'Reserved national' ,
    60 : 'Reserved national' ,
    61 : 'Reserved private' ,
    62 : 'Reserved private' ,
    63 : 'Reserved private' ,
    64 : 'Message authentication code (MAC)' ,
    65 : 'Bitmap, extended' ,
    66 : 'Settlement code' ,
    67 : 'Extended payment code' ,
    68 : 'Receiving institution country code' ,
    69 : 'Settlement institution country code' ,
    70 : 'Network management information code' ,
    71 : 'Message number' ,
    72 : 'Message number, last' ,
    73 : 'Date, action (YYMMDD)' ,
    74 : 'Credits, number' ,
    75 : 'Credits, reversal number' ,
    76 : 'Debits, number' ,
    77 : 'Debits, reversal number' ,
    78 : 'Transfer number' ,
    79 : 'Transfer, reversal number' ,
    80 : 'Inquiries number' ,
    81 : 'Authorizations, number' ,
    82 : 'Credits, processing fee amount' ,
    83 : 'Credits, transaction fee amount' ,
    84 : 'Debits, processing fee amount' ,
    85 : 'Debits, transaction fee amount' ,
    86 : 'Credits, amount' ,
    87 : 'Credits, reversal amount' ,
    88 : 'Debits, amount' ,
    89 : 'Debits, reversal amount' ,
    90 : 'Original data elements' ,
    91 : 'File update code' ,
    92 : 'File security code' ,
    93 : 'Response indicator' ,
    94 : 'Service indicator' ,
    95 : 'Replacement amounts' ,
    96 : 'Message security code' ,
    97 : 'Amount, net settlement' ,
    98 : 'Payee' ,
    99 : 'Settlement institution identification code' ,
    100 : 'Receiving institution identification code' ,
    101 : 'File name' ,
    102 : 'Account identification 1' ,
    103 : 'Account identification 2' ,
    104 : 'Transaction description' ,
    105 : 'Reserved for ISO use' ,
    106 : 'Reserved for ISO use' ,
    107 : 'Reserved for ISO use' ,
    108 : 'Reserved for ISO use' ,
    109 : 'Reserved for ISO use' ,
    110 : 'Reserved for ISO use' ,
    111 : 'Reserved for ISO use' ,
    112 : 'Reserved for national use' ,
    113 : 'Reserved for national use' ,
    114 : 'Reserved for national use' ,
    115 : 'Reserved for national use' ,
    116 : 'Reserved for national use' ,
    117 : 'Reserved for national use' ,
    118 : 'Reserved for national use' ,
    119 : 'Reserved for national use' ,
    120 : 'Reserved for private use' ,
    121 : 'Reserved for private use' ,
    122 : 'Reserved for private use' ,
    123 : 'Reserved for private use' ,
    124 : 'Reserved for private use' ,
    125 : 'Reserved for private use' ,
    126 : 'Reserved for private use' ,
    127 : 'Reserved for private use' ,
    128 : 'Message authentication code'
}
    
ContentTypes = {}

"""
Notes to the BPC's flavour of ISO8583:
 * DE 12 has length of 12, not 6
 * DE 39 has length of 3, not 2
"""
ContentTypes['1987'] = {
    1 :   { 'ContentType':'b',     'MaxLen': 8,   'LenType': LT.FIXED },
    2 :   { 'ContentType':'n',     'MaxLen': 19,  'LenType': LT.LLVAR },
    3 :   { 'ContentType':'n',     'MaxLen': 6,   'LenType': LT.FIXED },
    4 :   { 'ContentType':'n',     'MaxLen': 12,  'LenType': LT.FIXED },
    5 :   { 'ContentType':'n',     'MaxLen': 12,  'LenType': LT.FIXED },
    6 :   { 'ContentType':'n',     'MaxLen': 12,  'LenType': LT.FIXED },
    7 :   { 'ContentType':'n',     'MaxLen': 10,  'LenType': LT.FIXED },
    8 :   { 'ContentType':'n',     'MaxLen': 8,   'LenType': LT.FIXED },
    9 :   { 'ContentType':'n',     'MaxLen': 8,   'LenType': LT.FIXED },
    10 :  { 'ContentType':'n',     'MaxLen': 8,   'LenType': LT.FIXED },
    11 :  { 'ContentType':'n',     'MaxLen': 6,   'LenType': LT.FIXED },
    12 :  { 'ContentType':'n',     'MaxLen': 12,  'LenType': LT.FIXED },
    13 :  { 'ContentType':'n',     'MaxLen': 4,   'LenType': LT.FIXED },
    14 :  { 'ContentType':'n',     'MaxLen': 4,   'LenType': LT.FIXED },
    15 :  { 'ContentType':'n',     'MaxLen': 6,   'LenType': LT.FIXED },
    16 :  { 'ContentType':'n',     'MaxLen': 4,   'LenType': LT.FIXED },
    17 :  { 'ContentType':'n',     'MaxLen': 4,   'LenType': LT.FIXED },
    18 :  { 'ContentType':'n',     'MaxLen': 4,   'LenType': LT.FIXED },
    19 :  { 'ContentType':'n',     'MaxLen': 3,   'LenType': LT.FIXED },
    20 :  { 'ContentType':'n',     'MaxLen': 3,   'LenType': LT.FIXED },
    21 :  { 'ContentType':'n',     'MaxLen': 3,   'LenType': LT.FIXED },
    22 :  { 'ContentType':'n',     'MaxLen': 3,   'LenType': LT.FIXED },
    23 :  { 'ContentType':'n',     'MaxLen': 3,   'LenType': LT.FIXED },
    24 :  { 'ContentType':'n',     'MaxLen': 3,   'LenType': LT.FIXED },
    25 :  { 'ContentType':'n',     'MaxLen': 2,   'LenType': LT.FIXED },
    26 :  { 'ContentType':'n',     'MaxLen': 2,   'LenType': LT.FIXED },
    27 :  { 'ContentType':'n',     'MaxLen': 1,   'LenType': LT.FIXED },
    28 :  { 'ContentType':'an',    'MaxLen': 9,   'LenType': LT.FIXED },
    29 :  { 'ContentType':'an',    'MaxLen': 9,   'LenType': LT.FIXED },
    30 :  { 'ContentType':'an',    'MaxLen': 9,   'LenType': LT.FIXED },
    31 :  { 'ContentType':'an',    'MaxLen': 9,   'LenType': LT.FIXED },
    32 :  { 'ContentType':'n',     'MaxLen': 11,  'LenType': LT.LLVAR },
    33 :  { 'ContentType':'n',     'MaxLen': 11,  'LenType': LT.LLVAR },
    34 :  { 'ContentType':'ns',    'MaxLen': 28,  'LenType': LT.LLVAR },
    35 :  { 'ContentType':'z',     'MaxLen': 37,  'LenType': LT.LLVAR },
    36 :  { 'ContentType':'n',     'MaxLen': 104, 'LenType': LT.LLLVAR},
    37 :  { 'ContentType':'an',    'MaxLen': 12,  'LenType': LT.FIXED },
    38 :  { 'ContentType':'an',    'MaxLen': 6,   'LenType': LT.FIXED },
    39 :  { 'ContentType':'an',    'MaxLen': 3,   'LenType': LT.FIXED },
    40 :  { 'ContentType':'an',    'MaxLen': 3,   'LenType': LT.FIXED },
    41 :  { 'ContentType':'ans',   'MaxLen': 8,   'LenType': LT.FIXED },
    42 :  { 'ContentType':'ans',   'MaxLen': 15,  'LenType': LT.FIXED },
    43 :  { 'ContentType':'ans',   'MaxLen': 40,  'LenType': LT.FIXED },
    44 :  { 'ContentType':'an',    'MaxLen': 25,  'LenType': LT.LLVAR },
    45 :  { 'ContentType':'an',    'MaxLen': 76,  'LenType': LT.LLVAR },
    46 :  { 'ContentType':'an',    'MaxLen': 999, 'LenType': LT.LLLVAR},
    47 :  { 'ContentType':'an',    'MaxLen': 999, 'LenType': LT.LLLVAR},
    48 :  { 'ContentType':'an',    'MaxLen': 999, 'LenType': LT.LLLVAR},
    49 :  { 'ContentType':'an',    'MaxLen': 3,   'LenType': LT.FIXED },
    50 :  { 'ContentType':'an',    'MaxLen': 3,   'LenType': LT.FIXED },
    51 :  { 'ContentType':'an',    'MaxLen': 3,   'LenType': LT.FIXED },
    52 :  { 'ContentType':'b',     'MaxLen': 16,   'LenType': LT.FIXED },
    53 :  { 'ContentType':'n',     'MaxLen': 18,  'LenType': LT.FIXED },
    54 :  { 'ContentType':'an',    'MaxLen': 120, 'LenType': LT.LLLVAR},
    55 :  { 'ContentType':'ans',   'MaxLen': 999, 'LenType': LT.LLLVAR},
    56 :  { 'ContentType':'ans',   'MaxLen': 999, 'LenType': LT.LLLVAR},
    57 :  { 'ContentType':'ans',   'MaxLen': 999, 'LenType': LT.LLLVAR},
    58 :  { 'ContentType':'ans',   'MaxLen': 999, 'LenType': LT.LLLVAR},
    59 :  { 'ContentType':'ans',   'MaxLen': 999, 'LenType': LT.LLLVAR},
    60 :  { 'ContentType':'ans',   'MaxLen': 999, 'LenType': LT.LLLVAR},
    61 :  { 'ContentType':'ans',   'MaxLen': 999, 'LenType': LT.LLLVAR},
    62 :  { 'ContentType':'ans',   'MaxLen': 999, 'LenType': LT.LLLVAR},
    63 :  { 'ContentType':'ans',   'MaxLen': 999, 'LenType': LT.LLLVAR},
    64 :  { 'ContentType':'b',     'MaxLen' : 8,  'LenType': LT.FIXED },
    65 :  { 'ContentType' : 'b',   'MaxLen' : 1,  'LenType': LT.FIXED },
    66 :  { 'ContentType' : 'n',   'MaxLen' : 1,  'LenType': LT.FIXED },
    67 :  { 'ContentType' : 'n',   'MaxLen' : 2,  'LenType': LT.FIXED },
    68 :  { 'ContentType' : 'n',   'MaxLen' : 3,  'LenType': LT.FIXED },
    69 :  { 'ContentType' : 'n',   'MaxLen' : 3,  'LenType': LT.FIXED },
    70 :  { 'ContentType' : 'n',   'MaxLen' : 3,  'LenType': LT.FIXED },
    71 :  { 'ContentType' : 'n',   'MaxLen' : 4,  'LenType': LT.FIXED },
    72 :  { 'ContentType' : 'n',   'MaxLen' : 4,  'LenType': LT.FIXED },
    73 :  { 'ContentType' : 'n',   'MaxLen' : 6,  'LenType': LT.FIXED },
    74 :  { 'ContentType' : 'n',   'MaxLen' : 10, 'LenType': LT.FIXED },
    75 :  { 'ContentType' : 'n',   'MaxLen' : 10, 'LenType': LT.FIXED },
    76 :  { 'ContentType' : 'n',   'MaxLen' : 10, 'LenType': LT.FIXED },
    77 :  { 'ContentType' : 'n',   'MaxLen' : 10, 'LenType': LT.FIXED },
    78 :  { 'ContentType' : 'n',   'MaxLen' : 10, 'LenType': LT.FIXED },
    79 :  { 'ContentType' : 'n',   'MaxLen' : 10, 'LenType': LT.FIXED },
    80 :  { 'ContentType' : 'n',   'MaxLen' : 10, 'LenType': LT.FIXED },
    81 :  { 'ContentType' : 'n',   'MaxLen' : 10, 'LenType': LT.FIXED },
    82 :  { 'ContentType' : 'n',   'MaxLen' : 12, 'LenType': LT.FIXED },
    83 :  { 'ContentType' : 'n',   'MaxLen' : 12, 'LenType': LT.FIXED },
    84 :  { 'ContentType' : 'n',   'MaxLen' : 12, 'LenType': LT.FIXED },
    85 :  { 'ContentType' : 'n',   'MaxLen' : 12, 'LenType': LT.FIXED },
    86 :  { 'ContentType' : 'n',   'MaxLen' : 16, 'LenType': LT.FIXED },
    87 :  { 'ContentType' : 'n',   'MaxLen' : 16, 'LenType': LT.FIXED },
    88 :  { 'ContentType' : 'n',   'MaxLen' : 16, 'LenType': LT.FIXED },
    89 :  { 'ContentType' : 'n',   'MaxLen' : 16, 'LenType': LT.FIXED },
    90 :  { 'ContentType' : 'n',   'MaxLen' : 42, 'LenType': LT.FIXED },
    91 :  { 'ContentType' : 'an',  'MaxLen' : 1,  'LenType': LT.FIXED },
    92 :  { 'ContentType' : 'an',  'MaxLen' : 2,  'LenType': LT.FIXED },
    93 :  { 'ContentType' : 'an',  'MaxLen' : 5,  'LenType': LT.FIXED },
    94 :  { 'ContentType' : 'an',  'MaxLen' : 7,  'LenType': LT.FIXED },
    95 :  { 'ContentType' : 'an',  'MaxLen' : 42, 'LenType': LT.FIXED },
    96 :  { 'ContentType' : 'b',   'MaxLen' : 8,  'LenType': LT.FIXED },
    97 :  { 'ContentType' : 'an',  'MaxLen' : 17, 'LenType': LT.FIXED },
    98 :  { 'ContentType' : 'ans', 'MaxLen' : 25, 'LenType': LT.FIXED },
    99 :  { 'ContentType' : 'n',   'MaxLen' : 11, 'LenType': LT.LLVAR },
    100 : { 'ContentType' : 'n',   'MaxLen' : 11, 'LenType': LT.LLVAR },
    101 : { 'ContentType' : 'ans', 'MaxLen' : 17, 'LenType': LT.LLVAR },
    102 : { 'ContentType' : 'ans', 'MaxLen' : 28, 'LenType': LT.LLVAR },
    103 : { 'ContentType' : 'ans', 'MaxLen' : 28, 'LenType': LT.LLVAR },
    104 : { 'ContentType' : 'ans', 'MaxLen' : 100, 'LenType': LT.LLLVAR },
    105 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    106 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    107 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    108 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    109 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    110 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    111 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    112 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    113 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    114 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    115 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    116 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    117 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    118 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    119 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    120 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    121 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    122 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    123 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    124 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    125 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    126 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    127 : { 'ContentType' : 'ans', 'MaxLen' : 999, 'LenType': LT.LLLVAR },
    128 : { 'ContentType' : 'b',   'MaxLen' : 8,  'LenType': LT.FIXED }
}



import socket
import struct
import sys

from bpc8583.tools import get_datetime, get_random_hex
from tracetools.tracetools import trace
from pynblock.tools import raw2str, get_pinblock
from Crypto.Cipher import DES3


class Terminal:

    def __init__(self, host=None, port=None, id=None, merchant=None, master_key=None, terminal_key=None, show_keys=None):
        """
        Terminal initialization
        """
        self.pinblock_format = '01'

        # Host to connect to
        if host:
            self.host = host
        else:
            self.host = '127.0.0.1'

        # Port ot connect to
        if port:
            try:
                self.port = int(port)
            except ValueError:
                print('Invalid TCP port: {}'.format(arg))
                sys.exit()
        else:
            self.port = 1337

        # Terminal ID
        if id:
            self.terminal_id = id
        else:
            self.terminal_id = '10001337'

        # Merchant ID
        if merchant:
            self.merchant_id = merchant
        else:
            self.merchant_id = '999999999999001'

        self.currency = '643'
        self.country_code = '643'
        
        # Keys
        self.keyfile_name = '.terminalkey.cache'
        if master_key:
            self.master_key = bytes.fromhex(master_key)
        else:
            self.master_key = bytes.fromhex('CF7730DBA6CAC5E13C3FB45CAF8D71E1')    

        if terminal_key:
            self.terminal_key = bytes.fromhex(terminal_key)
        else:
            stored_key = self.get_stored_key()
            if stored_key:
                self.terminal_key = bytes.fromhex(stored_key)
            else:
                self.terminal_key = bytes.fromhex('FA9F90D49CB27B7D14A3FA9CCCFF6CB7')
                self.store_terminal_key(raw2str(self.terminal_key))

        self.tpk_cipher = DES3.new(self.terminal_key, DES3.MODE_ECB)
        self.tmk_cipher = DES3.new(self.master_key, DES3.MODE_ECB)

        self.show_keys = show_keys
        self.print_keys()


    def print_keys(self):
        if self.show_keys:
            print('  Master key: {}'.format(raw2str(self.master_key)))
            print('Terminal key: {}'.format(raw2str(self.terminal_key)))
        

    def connect(self):
        """
        """
        try:
            self.sock = None
            for res in socket.getaddrinfo(self.host, self.port, socket.AF_UNSPEC, socket.SOCK_STREAM):
                af, socktype, proto, canonname, sa = res
                self.sock = socket.socket(af, socktype, proto)
                self.sock.connect(sa)
        except OSError as msg:
            print('Error connecting to {}:{} - {}'.format(self.host, self.port, msg))
            sys.exit()
        print('Connected to {}:{}'.format(self.host, self.port))


    def send(self, data, show_trace=True):
        """
        """
        if show_trace:
            trace('>> {} bytes sent:'.format(len(data)), data)
        return self.sock.send(data)


    def recv(self, show_trace=True):
        """
        """
        data = self.sock.recv(4096)
        if show_trace:
            trace('<< {} bytes received: '.format(len(data)), data)
        return data

    def close(self):
        """
        """
        print('Disconnected from {}'.format(self.host))
        self.sock.close()


    def get_terminal_id(self):
        """
        """
        return self.terminal_id


    def get_merchant_id(self):
        """
        """
        return self.merchant_id


    def get_currency_code(self):
        """
        """
        return self.currency


    def store_terminal_key(self, key):
        """
        """
        try:
            f = open(self.keyfile_name, 'w+')
            f.write(key)
            f.close()
        except:
            return False
        return True


    def get_stored_key(self):
        """
        """
        try:
            f = open(self.keyfile_name, 'r')
            key = f.read()
            f.close()
            return key
        except:
            return None


    def set_terminal_key(self, encrypted_key):
        """
        Change the terminal key. The encrypted_key is a hex string
        encrypted_key is expected to be encrypted under master key
        """
        if encrypted_key:
            try:
                new_key = bytes.fromhex(encrypted_key)
                if len(self.terminal_key) != len(new_key):
                    # The keys must have equal length
                    return False

                self.terminal_key = self.tmk_cipher.decrypt(new_key)
                self.store_terminal_key(raw2str(self.terminal_key))

                self.tpk_cipher = DES3.new(self.terminal_key, DES3.MODE_ECB)
                self.print_keys()
                return True

            except ValueError:
                return False

        return False


    def get_terminal_key(self):
        """
        Get string representation of terminal key (needed mostly for debugging purposes)
        """
        return raw2str(self.terminal_key)


    def get_encrypted_pin(self, clear_pin, card_number):
        """
        Get PIN block in ISO 0 format, encrypted with the terminal key
        """
        if not self.terminal_key:
            print('Terminal key is not set')
            return ''

        if self.pinblock_format == '01':
            try:
                pinblock = bytes.fromhex(get_pinblock(clear_pin, card_number))
                #print('PIN block: {}'.format(raw2str(pinblock)))
            except TypeError:
                return ''

            encrypted_pinblock = self.tpk_cipher.encrypt(pinblock)
            return raw2str(encrypted_pinblock)

        else:
            print('Unsupported PIN Block format')
            return ''

    def get_pos_entry_mode(self):
        """
        """
        pan_and_date_entry_mode = '90'
        pin_entry_capability = '0'
        return int(pan_and_date_entry_mode + pin_entry_capability)


    def get_tvr(self):
        """
        return TVR (The status of the different functions as seen from the terminal)
        TODO: non-dummy value
        """
        return '0000040880'


    def get_unpredno(self):
        """
        Get unpredictable number (a value to provide variability and uniqueness to the generation of a cryptogram), used in ICC data
        """
        return get_random_hex(8)


    def get_country_code(self):
        """
        Get terminal country code, used in ICC data
        """
        return self.country_code.rjust(4, '0')



import binascii
import struct
from enum import Enum
from pytlv.TLV import TLV
from tracetools.tracetools import get_timestamp

# Length Type enumeration
class LT(Enum):
    FIXED   = 0
    LVAR    = 1
    LLVAR   = 2
    LLLVAR  = 3

# Data Type enumeration
class DT(Enum):
    BCD     = 1
    ASCII   = 2
    BIN     = 3

class SpecError(Exception):
        def __init__(self, value):
                self.str = value
        def __str__(self):
                return repr(self.str)

def MemDump(Title, data):
    i = 1

    if( isinstance(data, bytes) == False ):
        raise TypeError("Expected bytes for data")

    print(Title)
    TheDump = ""
    
    for c in data:
        try: # python 3
            TheDump += "{:02x} ".format(c) 
        except: # python 2.x
            TheDump += "{:02x} ".format(ord(c))
        
        if(i % 16 == 0):
            TheDump += "\n"
        i+=1
       
    print(TheDump)


def Bcd2Str(bcd):
    return binascii.hexlify(bcd).decode('latin')


def Str2Bcd(string):
    if(len(string) % 2 == 1):
        string = string.zfill(len(string) + 1)
    return binascii.unhexlify(string)


def Bcd2Int(bcd):
    return int(Bcd2Str(bcd))


def Int2Bcd(integer):
    string = str(integer)
    if(len(string) % 2 == 1):
        string = string.zfill(len(string) + 1)
    return binascii.unhexlify(string)


class ParseError(Exception):
        def __init__(self, value):
                self.str = value
        def __str__(self):
                return repr(self.str)
class SpecError(Exception):
        def __init__(self, value):
                self.str = value
        def __str__(self):
                return repr(self.str)
class BuildError(Exception):
        def __init__(self, value):
                self.str = value
        def __str__(self):
                return repr(self.str)
            
            
class ISO8583:
    
    ValidContentTypes = ('a', 'n', 's', 'an', 'as', 'ns', 'ans', 'b', 'z')
    
    def __init__(self, IsoMsg = None, IsoSpec = None):
        
        self.strict = False
    
        self.__Bitmap = {}
        self.__FieldData = {}
        self.__iso = b''
        self.tlv = TLV()
        
        if(IsoSpec != None):
            self.__IsoSpec = IsoSpec
        else:
            self.__IsoSpec = spec.IsoSpec1987ASCII()
        
        if(IsoMsg != None):
            if( isinstance(IsoMsg, bytes) == False ):
                raise TypeError("Expected bytes for iso message")
            
            self.__iso = IsoMsg
            self.ParseIso()


    def Strict(self, Value):
        if(Value != True and Value != False):
            raise ValueError
        self.strict = Value

        
    def SetIsoContent(self, IsoMsg):
        if( isinstance(IsoMsg, bytes) == False ):
            raise TypeError("Expected bytes for iso message")
        self.__iso = IsoMsg
        self.ParseIso()
    
    
    
    def ParseMTI(self, p):
        DataType = self.__IsoSpec.DataType('MTI')
        
        if(DataType == DT.BCD):
            self.__MTI = Bcd2Str(self.__iso[p:p+2])
            p+=2
        elif(DataType == DT.ASCII):
            self.__MTI = self.__iso[p:p+4].decode('latin')
            p+=4
        
        try: # MTI should only contain numbers
            int(self.__MTI)
        except:
            raise ParseError("Invalid MTI: [{0}]".format(self.__MTI))
            
        if(self.strict == True):
            if(self.__MTI[1] == '0'):
                raise ParseError("Invalid MTI: Invalid Message type [{0}]".format(self.__MTI))
                  
            if(int(self.__MTI[3]) > 5):
                raise ParseError("Invalid MTI: Invalid Message origin [{0}]".format(self.__MTI))
        
        return p
    
    
    def ParseBitmap(self, p):
        DataType = self.__IsoSpec.DataType(1)

        if(DataType == DT.BIN):
            Primary = self.__iso[p:p+8]
            p += 8
        elif(DataType == DT.ASCII):
            Primary = binascii.unhexlify(self.__iso[p:p+16])
            p += 16
                
        IntPrimary = struct.unpack_from("!Q", Primary)[0]
        
        for i in range(1, 65):
            self.__Bitmap[i] = (IntPrimary >> (64 - i)) & 0x1

        if(self.__Bitmap[1] == 1):
            if(DataType == DT.BIN):
                Secondary = self.__iso[p:p+8]
                p += 8
            elif(DataType == DT.ASCII):
                Secondary = binascii.unhexlify(self.__iso[p:p+16])
                p += 16
                
            IntSecondary = struct.unpack_from("!Q", Secondary)[0]
            
            for i in range(1, 65):
                self.__Bitmap[i+64] = (IntSecondary >> (64 - i)) & 0x1
            
        return p
            
            
    def ParseField(self, field, p):
        
        try:
            DataType = self.__IsoSpec.DataType(field)
            LenType = self.__IsoSpec.LengthType(field)
            ContentType = self.__IsoSpec.ContentType(field)
            MaxLength = self.__IsoSpec.MaxLength(field)
        except:
            raise SpecError("Cannot parse F{0}: Incomplete field specification".format(field))

        try:
            if(DataType == DT.ASCII and ContentType == 'b'):
                MaxLength *= 2
                
            if(LenType == LT.FIXED):
                Len = MaxLength
            elif(LenType == LT.LVAR):
                pass
            elif(LenType == LT.LLVAR):
                LenDataType = self.__IsoSpec.LengthDataType(field)
                if(LenDataType == DT.ASCII):
                    Len = int(self.__iso[p:p+2])
                    p+=2
                elif(LenDataType == DT.BCD):
                    Len = Bcd2Int(self.__iso[p:p+1])
                    p+=1
            elif(LenType == LT.LLLVAR):
                LenDataType = self.__IsoSpec.LengthDataType(field)
                if(LenDataType == DT.ASCII):
                    Len = int(self.__iso[p:p+3])
                    p+=3
                elif(LenDataType == DT.BCD):
                    Len = Bcd2Int(self.__iso[p:p+2])
                    p+=2
        except ValueError:
            raise ParseError("Cannot parse F{0}: Invalid length".format(field))
            
        if(Len > MaxLength):
            raise ParseError("F{0} is larger than maximum length ({1}>{2})".format(field, Len, MaxLength))
        
        # In case of zero length, don't try to parse the field itself, just continue
        if(Len == 0):
            return p

        try:
            if(DataType == DT.ASCII):
                if(ContentType == 'n'):
                    self.__FieldData[field] = int(self.__iso[p:p+(Len)])
                else:
                    self.__FieldData[field] = self.__iso[p:p+(Len)].decode('latin')
                p += Len
            elif(DataType == DT.BCD):
                if(Len % 2 == 1):
                    Len += 1
                if(ContentType == 'n'):
                    self.__FieldData[field] = Bcd2Int(self.__iso[p:p+(Len//2)])
                elif(ContentType == 'z'):
                    self.__FieldData[field] = binascii.hexlify(self.__iso[p:p+(Len//2)]).decode('latin').upper()
                p += Len//2
            elif(DataType == DT.BIN):
                self.__FieldData[field] = binascii.hexlify(self.__iso[p:p+(Len)]).decode('latin').upper()
                p += Len
        except:
            raise ParseError("Cannot parse F{0}".format(field))
        
        if(ContentType == 'z'):
            self.__FieldData[field] = self.__FieldData[field].replace("D", "=") # in track2, replace d with =  
            self.__FieldData[field] = self.__FieldData[field].replace("F", "") # in track2, remove trailing f
 
        return p
    
    
    def ParseIso(self):
        p = 0
        p = self.ParseMTI(p)
        p = self.ParseBitmap(p)


        for field in sorted(self.__Bitmap):
            # field 1 is parsed by the bitmap function
            if(field != 1 and self.Field(field) == 1):
                p = self.ParseField(field, p)
    

    def BuildMTI(self):
        if(self.__IsoSpec.DataType('MTI') == DT.BCD):
            self.__iso += Str2Bcd(self.__MTI)
        elif(self.__IsoSpec.DataType('MTI') == DT.ASCII):
            self.__iso += self.__MTI.encode('latin')
    
    
    def BuildBitmap(self):
        DataType = self.__IsoSpec.DataType(1)
        
        # check if we need a secondary bitmap
        for i in self.__Bitmap.keys():
            if(i > 64):
                self.__Bitmap[1] = 1
                break
        
        IntPrimary = 0
        for i in range(1, 65):
            if i in self.__Bitmap.keys():
                IntPrimary |= (self.__Bitmap[i] & 0x1) << (64 - i)

        Primary = struct.pack("!Q", IntPrimary)

        if(DataType == DT.BIN):
            self.__iso += Primary
        elif(DataType == DT.ASCII):
            self.__iso += binascii.hexlify(Primary)
            
        # Add secondary bitmap if applicable
        if 1 in self.__Bitmap.keys() and self.__Bitmap[1] == 1:
        
            IntSecondary = 0
            for i in range(65, 129):
                if i in self.__Bitmap.keys():
                    IntSecondary |= (self.__Bitmap[i] & 0x1) << (128 - i)
                
            Secondary = struct.pack("!Q", IntSecondary)

            if(DataType == DT.BIN):
                self.__iso += Secondary
            elif(DataType == DT.ASCII):
                self.__iso += binascii.hexlify(Secondary)
            
            
    def BuildField(self, field):
        try:
            DataType = self.__IsoSpec.DataType(field)
            LenType = self.__IsoSpec.LengthType(field)
            ContentType = self.__IsoSpec.ContentType(field)
            MaxLength = self.__IsoSpec.MaxLength(field)
        except:
            raise SpecError("Cannot parse F{0}: Incomplete field specification".format(field))
 

        data = ""
        if(LenType == LT.FIXED):
            Len = MaxLength
            
            if(ContentType == 'n'):
                formatter = "{{0:0{0}d}}".format(Len)
            elif('a' in ContentType or 'n' in ContentType or 's' in ContentType):
                formatter = "{{0: >{0}}}".format(Len)
            else:
                formatter = "{0}"
                
            data = formatter.format(self.__FieldData[field])
                
        else:
            LenDataType = self.__IsoSpec.LengthDataType(field)
            
            try:
                data = '{0}'.format(self.__FieldData[field])
            except KeyError:
                data = ''
            Len = len(data)
            if(DataType == DT.BIN):
                Len //=2
                
            if(Len > MaxLength):
                raise BuildError("Cannot Build F{0}: Field Length larger than specification".format(field))
            
            if(LenType == LT.LVAR):
                LenData = "{0:01d}".format(Len)
                
            elif(LenType == LT.LLVAR):
                LenData = "{0:02d}".format(Len)
                
            elif(LenType == LT.LLLVAR):
                LenData = "{0:03d}".format(Len)
                
            if(LenDataType == DT.ASCII):
                self.__iso += LenData.encode('latin')
            elif(LenDataType == DT.BCD):
                self.__iso += Str2Bcd(LenData)
            elif(LenDataType == DT.BIN):
                self.__iso += binascii.unhexlify(LenData)
            
            
        if(ContentType == 'z'):
            data = data.replace("=", "D")
            #if(len(data) % 2 == 1):
            #    data = data + 'F'
        
        if(DataType == DT.ASCII):
            self.__iso += data.encode('latin')
        elif(DataType == DT.BCD):
            self.__iso += Str2Bcd(data)
        elif(DataType == DT.BIN):
            self.__iso += binascii.unhexlify(self.__FieldData[field])


    def BuildIso(self):
        self.__iso = b''
        self.BuildMTI()
        self.BuildBitmap()
        
        for field in sorted(self.__Bitmap):
            if(field != 1 and self.Field(field) == 1):
                self.BuildField(field)
                
        return self.__iso
    
    
    def RemoveField(self, field):
        """
        """
        try:
            self.__FieldData[field] = None
            del(self.__Bitmap[field])
        except KeyError:
            pass


    def Field(self, field, Value = None):
        """
        Add field to bitmap
        """
        if(Value == None):
            try:
                return self.__Bitmap[field]
            except KeyError:
                return None
        elif(Value == 1 or Value == 0):
            self.__Bitmap[field] = Value
        else:
            raise ValueError


    def SetBitmap(self, fields):
        """
        Set the message bitmap with the value from fields array
        """
        for field in fields:
            self.Field(field, Value=1)
            

    def FieldData(self, field, Value = None):
        """
        Add field data
        """
        if(Value == None):
            try:
                return self.__FieldData[field]
            except KeyError:
                return None
        else:
            if(len(str(Value)) > self.__IsoSpec.MaxLength(field)):
                raise ValueError('Value length larger than field maximum ({0})'.format(self.__IsoSpec.MaxLength(field)))
            
            self.Field(field, Value=1)
            self.__FieldData[field] = Value 
            
            
    def Bitmap(self):
        return self.__Bitmap

    def MTI(self, MTI = None):
        if(MTI == None):
            return self.__MTI
        else:
            try: # MTI should only contain numbers
                int(MTI)
            except:
                raise ValueError("Invalid MTI [{0}]: MTI must contain only numbers".format(MTI))
        
            if(self.strict == True):
                if(MTI[1] == '0'):
                    raise ValueError("Invalid MTI [{0}]: Invalid Message type".format(MTI))
                      
                if(int(MTI[3]) > 5):
                    raise ValueError("Invalid MTI [{0}]: Invalid Message origin".format(MTI))
            
            self.__MTI = MTI


    def get_MTI(self):
        """
        """
        return self.__MTI


    def Description(self, field, Description = None):
        return self.__IsoSpec.Description(field, Description)
    
    def DataType(self, field, DataType = None):
        return self.__IsoSpec.DataType(field, DataType)
    
    def ContentType(self, field, ContentType = None):
        return self.__IsoSpec.ContentType(field, ContentType)


    def PrintMessage(self):
        self.Print()


    def Print(self, header=None):
        if header:
            print('\t{} at {}:'.format(header, get_timestamp()))
        else:
            print('\tParsed message:')

        try:
            print("\tMTI:    [{0}]".format(self.__MTI))
        except AttributeError:
            pass
        
        bitmapLine = "\tFields: [ "
        for i in sorted(self.__Bitmap.keys()):
            if(i == 1): 
                continue
            if(self.__Bitmap[i] == 1):
                bitmapLine += str(i) + " "
        bitmapLine += "]"
        print(bitmapLine)
        

        for i in sorted(self.__Bitmap.keys()):
            if(i == 1): 
                continue
            if(self.__Bitmap[i] == 1):
                
                try:
                    FieldData = self.__FieldData[i]
                except KeyError:
                    FieldData = ''
                
                if(self.ContentType(i) == 'n' and self.__IsoSpec.LengthType(i) == LT.FIXED):
                    FieldData = str(FieldData).zfill(self.__IsoSpec.MaxLength(i))
                if i == 39:
                    print("\t\t{0:>3d} - {1: <41} [{2}]\t\t\t[{3}]".format(i, self.__IsoSpec.Description(i), FieldData, self.__IsoSpec.RespCodeDescription(FieldData)))
                elif i == 55:
                    print("\t\t{0:>3d} - {1: <41} [{2}]".format(i, self.__IsoSpec.Description(i), FieldData))
                    print(self.tlv.dump(self.tlv.parse(FieldData), left_indent='\t\t', desc_column_width=38))
                else:
                    print("\t\t{0:>3d} - {1: <41} [{2}]".format(i, self.__IsoSpec.Description(i), FieldData))

        print('\n')

currency_codes = {
'AFA': 4,
'ALL': 8,
'DZD': 12,
'ADP': 20,
'AOK': 24,
'ARA': 32,
'AUD': 36,
'ATS': 40,
'BSD': 44,
'BHD': 48,
'BDT': 50,
'AMD': 51,
'BBD': 52,
'BEF': 56,
'BMD': 60,
'BTN': 64,
'BOB': 68,
'BWP': 72,
'BRC': 76,
'BZD': 84,
'SBD': 90,
'BND': 96,
'BGL': 100,
'MMK': 104,
'BIF': 108,
'KHR': 116,
'CAD': 124,
'CVE': 132,
'KYD': 136,
'LKR': 144,
'CLP': 152,
'CNY': 156,
'COP': 170,
'KMF': 174,
'ZRZ': 180,
'CRC': 188,
'HRK': 191,
'CUP': 192,
'CYP': 196,
'CSK': 200,
'CZK': 203,
'DKK': 208,
'DOP': 214,
'ECS': 218,
'SVC': 222,
'GQE': 226,
'ETB': 230,
'EEK': 233,
'FKP': 238,
'FJD': 242,
'FIM': 246,
'FRF': 250,
'DJF': 262,
'GMD': 270,
'DDM': 278,
'DEM': 280,
'GHC': 288,
'GIP': 292,
'GRD': 300,
'GTQ': 320,
'GNF': 324,
'GYD': 328,
'HTG': 332,
'HNL': 340,
'HKD': 344,
'HUF': 348,
'ISK': 352,
'INR': 356,
'IDR': 360,
'IRR': 364,
'IRA': 365,
'IQD': 368,
'IEP': 372,
'ILS': 376,
'ITL': 380,
'JMD': 388,
'JPY': 392,
'KZT': 398,
'JOD': 400,
'KES': 404,
'KPW': 408,
'KRW': 410,
'KWD': 414,
'KGS': 417,
'LAK': 418,
'LBP': 422,
'LSL': 426,
'LVL': 428,
'LRD': 430,
'LYD': 434,
'LTL': 440,
'LUF': 442,
'MOP': 446,
'MGF': 450,
'MWK': 454,
'MYR': 458,
'MVR': 462,
'MLF': 466,
'MTL': 470,
'MRO': 478,
'MUR': 480,
'MXN': 484,
'MNT': 496,
'MDL': 498,
'MAD': 504,
'MZM': 508,
'OMR': 512,
'NAD': 516,
'NPR': 524,
'NLG': 528,
'ANG': 532,
'AWG': 533,
'VUV': 548,
'NZD': 554,
'NIO': 558,
'NGN': 566,
'NOK': 578,
'PKR': 586,
'PAB': 590,
'PGK': 598,
'PYG': 600,
'PEI': 604,
'PHP': 608,
'PLN': 616,
'PTE': 620,
'GWP': 624,
'TPE': 626,
'QAR': 634,
'ROL': 642,
'RUB': 643,
'RWF': 646,
'SHP': 654,
'STD': 678,
'SAR': 682,
'SCR': 690,
'SLL': 694,
'SGD': 702,
'SKK': 703,
'VND': 704,
'SIT': 705,
'SOS': 706,
'ZAR': 710,
'ZWD': 716,
'YDD': 720,
'ESP': 724,
'SSP': 728,
'SDD': 736,
'SDA': 737,
'SRG': 740,
'SZL': 748,
'SEK': 752,
'CHF': 756,
'SYP': 760,
'THB': 764,
'TOP': 776,
'TLP': 777,
'TTD': 780,
'AED': 784,
'TND': 788,
'TRL': 792,
'UGX': 800,
'UAK': 804,
'MKD': 807,
'SUR': 810,
'EGP': 818,
'GBP': 826,
'TZS': 834,
'USD': 840,
'UYP': 858,
'UZS': 860,
'VEB': 862,
'WST': 882,
'YER': 886,
'YUD': 890,
'ZMK': 894,
'TWD': 901,
'BYN': 933,
'TMT': 934,
'SDG': 938,
'AZN': 944,
'TRY': 949,
'XAF': 950,
'XCD': 951,
'XOF': 952,
'XPF': 953,
'ZMW': 967,
'SRD': 968,
'MGA': 969,
'TJS': 972,
'BYR': 974,
'BAM': 977,
'EUR': 978,
'GEL': 981,
'BRL': 986,
'BEL': 992,
'BEC': 993,
'XXX': 999}


from setuptools import setup

setup(name='bpc8583',
      version='1.39',
      
      description='ISO8583 library and tools (BPC\'s flavour)',
      long_description=open('README').read(),
      
      classifiers=[
        'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'Operating System :: OS Independent',
        
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        
        'Topic :: Communications',
        'Intended Audience :: Developers',
      ],
      
      keywords='POS ISO8583 banking financial payment',
      
      url='https://github.com/timgabets/bpc8583',
      author='Tim Gabets',
      author_email='tim@gabets.ru',
      
      license='LGPLv2',
      packages=['bpc8583'],
      install_requires=['enum34', 'pycrypto', 'pytlv', 'tracetools'],
      zip_safe=True)


import random

from time import strftime
from math import pow
from datetime import datetime
from tracetools.tracetools import get_timestamp


def get_date():
    """
    Get current date YYMMDD
    """
    return datetime.now().strftime("%y%m%d")


def get_MMDD():
    """
    Get current date MMDD
    """
    return int(datetime.now().strftime("%m%d"))


def get_datetime():
    """
    Get current date and time (for ISO data fields)
    """
    return int(datetime.now().strftime("%d%m%H%M%S"))


def get_datetime_with_year():
    """
    Get current time (for ISO data fields)
    """
    return int(datetime.now().strftime("%y%m%d%H%M%S"))


def get_seconds_since_epoch():
    """
    Get number of seconds since Jan 01 1970
    """
    return int((datetime.now()-datetime(1970,1,1)).total_seconds())


def get_stan():
    """
    Get random systems trace audit number
    """
    return random.randint(0, 999999)


def hexify(number):
    """
    Convert integer to hex string representation, e.g. 12 to '0C'
    """
    if number < 0:
        raise ValueError('Invalid number to hexify - must be positive')

    result = hex(int(number)).replace('0x', '').upper()
    if divmod(len(result), 2)[1] == 1:
        # Padding
        result = '0{}'.format(result)
    return result


def get_random_hex(length):
    """
    Return random hex string of a given length
    """
    if length <= 0:
        return ''
    return hexify(random.randint(pow(2, length*2), pow(2, length*4)))[0:length]


def get_response(_code):
    """
    Return xx1x response for xx0x codes (e.g. 0810 for 0800)
    """
    if _code:
        code = str(_code)
        return code[:-2] + str(int(code[-2:-1]) + 1) + code[-1]
    else:
        return None


def trace_passed(description, show_colored_description=False):
    """
    """
    padding = ''
    for i in range(56 - len(description)):
        padding += ' '
    if show_colored_description:
        print("{}\t| \033[32m{}\033[0m{}[\033[32mPASSED\033[0m]".format(get_timestamp(), description[:56], padding))
    else:
        print("{}\t| {}{}[\033[32mPASSED\033[0m]".format(get_timestamp(), description[:56], padding))


def trace_failed(description, actual_response, show_colored_description=False):
    """
    """
    padding = ''
    for i in range(56 - len(description)):
        padding += ' '
    if show_colored_description:
        print('{}\t| \033[31m{}\033[0m{}[\033[31mFAILED\033[0m]'.format(get_timestamp(), description[:56], padding))
    else:
        print('{}\t| {}{}[\033[31mFAILED\033[0m]'.format(get_timestamp(), description[:56], padding))
